#cython: language_level=3str

from libc.stdint cimport int64_t

from enum import Enum


import warnings

import numpy as np
import scipy as sp

from .csc cimport _CSCMatrix
from .csc import _CSCMatrix


class EfficiencyWarning:
    pass


cdef extern from "spral.h":

    struct spral_ssids_options:
        int array_base
        int print_level
        int unit_diagnostics
        int unit_error
        int unit_warning
        int ordering
        int nemin
        bint ignore_numa
        bint use_gpu
        int64_t min_gpu_work
        float max_load_inbalance
        float gpu_perf_coeff
        int scaling
        int64_t small_subtree_threshold
        int cpu_block_size
        bint action
        int pivot_method
        double small
        double u
        char unused[80]

    struct spral_ssids_inform:
        int flag
        int matrix_dup
        int matrix_missing_diag
        int matrix_outrange
        int matrix_rank
        int maxdepth
        int maxfront
        int num_delay
        int64_t num_factor
        int64_t num_flops
        int num_neg
        int num_sup
        int num_two
        int stat
        int cuda_error
        int cublas_error
        int maxsupernode
        char unused[76]

    void spral_ssids_default_options(spral_ssids_options *options)

    void spral_ssids_analyse_ptr32(bint check, int n, int *order, const int *ptr,
                                   const int *row, const double *val, void **akeep,
                                   const spral_ssids_options *options,
                                   spral_ssids_inform *inform)

    void spral_ssids_factor_ptr32(bint posdef, const int *ptr, const int *row,
                                  const double *val, double *scale, void *akeep, void **fkeep,
                                  const spral_ssids_options *options,
                                  spral_ssids_inform *inform)

    void spral_ssids_solve(int job, int nrhs, double *x, int ldx, void *akeep,
                           void *fkeep, const spral_ssids_options *options,
                           spral_ssids_inform *inform)

    void spral_ssids_solve1(int job, double *x1, void *akeep, void *fkeep,
                            const spral_ssids_options *options,
                            spral_ssids_inform *inform)

    int spral_ssids_free_akeep(void **akeep)
    int spral_ssids_free_fkeep(void **fkeep)
    int spral_ssids_free(void **akeep, void **fkeep)

    void spral_ssids_enquire_posdef(const void *akeep, const void *fkeep,
                                    const spral_ssids_options *options,
                                    spral_ssids_inform *inform, double *d)

    void spral_ssids_enquire_indef(const void *akeep, const void *fkeep,
                                   const spral_ssids_options *options,
                                   spral_ssids_inform *inform, int *piv_order, double *d)

    void spral_ssids_alter(const double *d, const void *akeep, void *fkeep,
                           const spral_ssids_options *options,
                           spral_ssids_inform *inform)


ErrorFlags = {
    0: "Success.",
    # Errors
    -1: "Error in sequence of calls (may be caused by failure of a preceding call).",
    -2: "n<0 or ne<1.",
    -3: "Error in ptr(:).",
    -4: "CSC format: All variable indices in one or more columns are out-of-range. Coordinate format: All entries are out-of-range.",
    -5: "Matrix is singular and options%action=.false.",
    -6: "Matrix found not to be positive definite but posdef=true.",
    -7: "ptr(:) and/or row(:) not present, but required as ssids_analyse() was called with check=.false,.",
    -8: "options%ordering out of range, or options%ordering=0 and order parameter not provided or not a valid permutation.",
    -9: "options%ordering=-2 but val(:) was not supplied.",
    -10: "ldx<n or nrhs<1.",
    -11: "job is out-of-range.",
    -13: "Called ssids_enquire_posdef() on indefinite factorization.",
    -14: "Called ssids_enquire_indef() on positive-definite factorization.",
    -15: "options%scaling=3 but a matching-based ordering was not performed during analyse phase.",
    -50: "Allocation error. If available, the stat parameter is returned in inform%stat.",
    -51: "CUDA error. The CUDA error return value is returned in inform%cuda_error.",
    -52: "CUBLAS error. The CUBLAS error return value is returned in inform%cublas_error.",
    -53: "OpenMP cancellation is disabled. Please set the environment variable OMP_CANCELLATION=true.",
    # Warnings
    1: "Out-of-range variable indices found and ignored in input data. inform%matrix_outrange is set to the number of such entries.",
    2: "Duplicate entries found and summed in input data. inform%matrix_dup is set to the number of such entries.",
    3: "Combination of +1 and +2.",
    4: "One or more diagonal entries of A are missing.",
    5: "Combination of +4 and +1 or +2.",
    6: "Matrix is found be (structurally) singular during analyse phase. This will overwrite any of the above warning flags.",
    7: "Matrix is found to be singular during factorize phase.",
    8: "Matching-based scaling found as side-effect of matching-based ordering ignored (consider setting options%scaling=3).",
    50: "OpenMP processor binding is disabled. Consider setting the environment variable OMP_PROC_BIND=true (this may affect performance on NUMA systems)."
}


class PivotMethod(Enum):
    AgressiveAPosteriori = 1
    BlockAPosteriori = 2
    ThresholdPartial = 3


class Ordering(Enum):
    UserSupplied = 0
    METIS = 1
    MatchingBased = 2


class Scaling(Enum):
    NoScaling = 0
    MC64 = 1
    Auction = 2
    MC77 = 3
    NormEquilibration = 4


class PrintLevel(Enum):
    NoPrinting = -1
    ErrorWarning = 0
    BasicDiagnostic = 1
    DetailedDiagnostic = 2


cdef class _Inform:
    cdef spral_ssids_inform inform

    @property
    def flag(self):
        return self.inform.flag

    @property
    def matrix_dup(self):
        return self.inform.matrix_dup

    @property
    def matrix_missing_diag(self):
        return self.inform.matrix_missing_diag

    @property
    def matrix_outrange(self):
        return self.inform.matrix_outrange

    @property
    def matrix_rank(self):
        return self.inform.matrix_rank

    @property
    def maxdepth(self):
        return self.inform.maxdepth

    @property
    def maxfront(self):
        return self.inform.maxfront

    @property
    def num_delay(self):
        return self.inform.num_delay

    @property
    def num_factor(self):
        return self.inform.num_factor

    @property
    def num_flops(self):
        return self.inform.num_flops

    @property
    def num_neg(self):
        return self.inform.num_neg

    @property
    def num_sup(self):
        return self.inform.num_sup

    @property
    def num_two(self):
        return self.inform.num_two

    @property
    def stat(self):
        return self.inform.stat

    @property
    def cuda_error(self):
        return self.inform.cuda_error

    @property
    def cublas_error(self):
        return self.inform.cublas_error

    @property
    def maxsupernode(self):
        return self.inform.maxsupernode


cdef class Options:
    cdef spral_ssids_options options

    def __cinit__(self, **kwds):
        spral_ssids_default_options(&self.options)

    def __init__(self, **kwds):
        for key, value in kwds.items():
            setattr(self, key, value)

    property array_base:
        def __get__(self): return self.options.array_base
        def __set__(self, value): self.options.array_base = value

    property print_level:
        def __get__(self): return PrintLevel(self.options.print_level)
        def __set__(self, value): self.options.print_level = value.value

    property unit_diagnostics:
        def __get__(self): return self.options.unit_diagnostics
        def __set__(self, value): self.options.unit_diagnostics = value

    property unit_error:
        def __get__(self): return self.options.unit_error
        def __set__(self, value): self.options.unit_error = value

    property unit_warning:
        def __get__(self): return self.options.unit_warning
        def __set__(self, value): self.options.unit_warning = value

    property ordering:
        def __get__(self): return Ordering(self.options.ordering)
        def __set__(self, value): self.options.ordering = value.value

    property nemin:
        def __get__(self): return self.options.nemin
        def __set__(self, value): self.options.nemin = value

    property ignore_numa:
        def __get__(self): return self.options.ignore_numa
        def __set__(self, value): self.options.ignore_numa = value

    property use_gpu:
        def __get__(self): return self.options.use_gpu
        def __set__(self, value): self.options.use_gpu = value

    property min_gpu_work:
        def __get__(self): return self.options.min_gpu_work
        def __set__(self, value): self.options.min_gpu_work = value

    property max_load_imbalance:
        def __get__(self): return self.options.max_load_imbalance
        def __set__(self, value): self.options.max_load_imbalance = value

    property gpu_perf_coeff:
        def __get__(self): return self.options.gpu_perf_coeff
        def __set__(self, value): self.options.gpu_perf_coeff = value

    @property
    def scaling(self):
        scaling = min(max(self.options.scaling, 0), 4)
        return Scaling(scaling)

    @scaling.setter
    def scaling(self, value):
        self.options.scaling = value.value

    property small_subtree_threshold:
        def __get__(self): return self.options.small_subtree_threshold
        def __set__(self, value): self.options.small_subtree_threshold = value

    property cpu_block_size:
        def __get__(self): return self.options.cpu_block_size
        def __set__(self, value): self.options.cpu_block_size = value

    property action:
        def __get__(self): return self.options.action
        def __set__(self, value): self.options.action = value

    property pivot_method:
        def __get__(self): return PivotMethod(self.options.pivot_method)
        def __set__(self, value): self.options.pivot_method = value.value

    property small:
        def __get__(self): return self.options.small
        def __set__(self, value): self.options.small = value

    property u:
        def __get__(self): return self.options.u
        def __set__(self, value): self.options.u = value


cdef class SymbolicFactor:
    cdef void *akeep
    cdef _CSCMatrix csc_mat
    cdef Options options

    def __cinit__(self):
        self.akeep = NULL
        self.csc_mat = None
        self.options = None

    def __dealloc__(self):
        spral_ssids_free_akeep(&self.akeep)

    def factor(self, bint posdef, double[:] scale=None):
        cdef void* fkeep = NULL
        cdef double* scale_ptr = NULL
        cdef NumericFactor numeric_factor = NumericFactor()

        if scale is not None:
            if self.options.scaling != Scaling.NoScaling:
                raise ValueError("User-provided scaling with automatic scaling option")
            assert scale.ndim == 1
            (_, n) = self.csc_mat.shape
            assert scale.size == n
            scale_ptr = &scale[0]

        spral_ssids_factor_ptr32(posdef,
                                 &self.csc_mat.cols[0],
                                 &self.csc_mat.rows[0],
                                 &self.csc_mat.data[0],
                                 scale_ptr,
                                 self.akeep,
                                 &fkeep,
                                 &self.options.options,
                                 &numeric_factor._inform.inform)

        if numeric_factor._inform.inform.flag < 0:
            spral_ssids_free_fkeep(&fkeep)
            raise SSIDSError(numeric_factor._inform.inform.flag)

        numeric_factor.fkeep = fkeep
        numeric_factor.csc_mat = self.csc_mat
        numeric_factor.symbolic_factor = self
        numeric_factor.posdef = posdef

        return numeric_factor


class SSIDSError(Exception):
    def __init__(self, flag):
        self.flag = flag
        self.message = ErrorFlags.get(flag, "Unknown error")

    def __str__(self):
        return self.message


class Job(Enum):
    Axb = 0
    PLxSb = 1
    Dxb = 2
    PLTSinvxb = 3
    DPLTSinvxb = 4


cdef class NumericFactor:
    cdef _CSCMatrix csc_mat
    cdef SymbolicFactor symbolic_factor
    cdef bint posdef
    cdef void *fkeep
    cdef _Inform _inform

    def __cinit__(self):
        self.fkeep = NULL
        self._inform = _Inform()

    def __dealloc__(self):
        spral_ssids_free_fkeep(&self.fkeep)

    @property
    def inform(self):
        return self._inform

    cdef _solve_single(self, double[:] rhs, job=Job.Axb, inplace=False):
        cdef int _job = job.value
        cdef double[:] sol_rhs

        (m, _) = self.csc_mat.shape

        if rhs.shape[0] != m:
            raise ValueError("Invalid shape of rhs")

        if not inplace:
            sol_rhs = np.copy(rhs)
        else:
            sol_rhs = rhs

        spral_ssids_solve1(_job,
                           &sol_rhs[0],
                           self.symbolic_factor.akeep,
                           self.fkeep,
                           &self.symbolic_factor.options.options,
                           &self._inform.inform)

        if self._inform.inform.flag < 0:
            raise SSIDSError(self._inform.inform.flag)

        return sol_rhs

    cdef _solve_multi(self, double[:, ::1] rhs, job=Job.Axb, inplace=False):
        cdef int _job = job.value
        cdef double[:, ::1] sol_rhs
        cdef int nrhs
        cdef int leadim

        (m, _) = self.csc_mat.shape

        if rhs.shape[1] != m:
            raise ValueError("Invalid shape of rhs")

        if not inplace:
            sol_rhs = np.copy(rhs)
        else:
            sol_rhs = rhs

        nrhs = sol_rhs.shape[0]
        leadim = sol_rhs.strides[0] // sol_rhs.itemsize

        spral_ssids_solve(_job,
                          nrhs,
                          &sol_rhs[0, 0],
                          leadim,
                          self.symbolic_factor.akeep,
                          self.fkeep,
                          &self.symbolic_factor.options.options,
                          &self._inform.inform)

        if self._inform.inform.flag < 0:
            raise SSIDSError(self._inform.inform.flag)

        return np.asarray(sol_rhs)

    def solve(self, rhs, job=Job.Axb, inplace=False):

        if rhs.ndim == 1:
            return self._solve_single(rhs, job, inplace)
        else:
            return self._solve_multi(rhs, job, inplace)

    cdef _enquire_posdef(self):
        cdef double[:] d

        (_, n) = self.csc_mat.shape
        d = np.empty(n, dtype=np.double)

        spral_ssids_enquire_posdef(self.symbolic_factor.akeep,
                                   self.fkeep,
                                   &self.symbolic_factor.options.options,
                                   &self._inform.inform,
                                   &d[0])

        if self._inform.inform.flag < 0:
            raise SSIDSError(self._inform.inform.flag)

        return d

    cdef _enquire_indef(self):
        cdef double[:, ::1] d
        cdef int[:] piv_order

        (_, n) = self.csc_mat.shape

        d_arr = np.empty(shape=(n, 2), dtype=np.double)
        d = d_arr
        piv_order_arr = np.empty(n, dtype=np.intc)
        piv_order = piv_order_arr

        spral_ssids_enquire_indef(self.symbolic_factor.akeep,
                                 self.fkeep,
                                 &self.symbolic_factor.options.options,
                                 &self._inform.inform,
                                 &piv_order[0],
                                 &d[0, 0])

        if self._inform.inform.flag < 0:
            raise SSIDSError(self._inform.inform.flag)

        return piv_order_arr, d_arr

    def alter(self, double[:, ::1] d):
        (_, n) = self.csc_mat.shape

        if (d.shape[0] != n) or (d.shape[1] != 2):
            raise ValueError("Invalid shape of d")

        spral_ssids_alter(&d[0, 0],
                          self.symbolic_factor.akeep,
                          self.fkeep,
                          &self.symbolic_factor.options.options,
                          &self._inform.inform)

        if self._inform.inform.flag < 0:
            raise SSIDSError(self._inform.inform.flag)

    def D(self):
        if self.posdef:
            d = self._enquire_posdef()
        else:
            piv_order, d = self._enquire_indef()
            piv_perm = np.abs(piv_order) - 1
            d = d[piv_perm, :]

        (_, n) = self.csc_mat.shape

        dsub = d[1:, 1]
        dmain = d[:, 0]

        diagonals = [dsub, dmain, dsub]
        return sp.sparse.diags_array(diagonals, offsets=[-1, 0, 1])

    def enquire(self):
        if self.posdef:
            return self._enquire_posdef()
        else:
            return self._enquire_indef()


cdef _analyze(mat, Options options, check=False, int[:] order=None):
    cdef void* akeep = NULL
    cdef void* fkeep = NULL
    cdef int* order_ptr = NULL
    cdef const spral_ssids_options* _options
    cdef _CSCMatrix csc_mat
    cdef SymbolicFactor symbolic_factor = SymbolicFactor()
    cdef spral_ssids_inform inform

    assert options is not None
    options.array_base = 1

    if not sp.sparse.issparse(mat):
        raise ValueError("Input matrix must be a sparse matrix")

    if mat.format != 'csc':
        warnings.warn("Input matrix is not in CSC format. Converting to CSC format.", EfficiencyWarning)
        mat = mat.tocsc()

    (m, n) = mat.shape

    _options = &options.options
    csc_mat = _CSCMatrix(mat, symmetric=True)

    if m != n:
        raise ValueError("Input matrix must be square")

    if order is not None:
        if options.ordering != Ordering.UserSupplied:
            raise ValueError("User-provided ordering with automatic ordering option")
        assert order.ndim == 1
        assert order.size == n
        order_ptr = &order[0]

    if mat.dtype != float:
        raise ValueError("Input matrix must be of type float")

    spral_ssids_analyse_ptr32(check,
                              n,
                              order_ptr,
                              &csc_mat.cols[0],
                              &csc_mat.rows[0],
                              &csc_mat.data[0],
                              &akeep,
                              _options,
                              &inform)

    if inform.flag < 0:
        spral_ssids_free(&akeep, &fkeep)
        raise SSIDSError(inform.flag)

    symbolic_factor.akeep = akeep
    symbolic_factor.csc_mat = csc_mat
    symbolic_factor.options = options

    return symbolic_factor


def analyze(mat, check=False, int[:] order=None, **options):
    cdef Options _options = Options(**options)
    return _analyze(mat, options=_options, check=check, order=order)


def solve(mat, rhs, posdef, check=False, inplace=False, **options):
    symbolic_factor = analyze(mat, check=check, **options)
    numeric_factor = symbolic_factor.factor(posdef=posdef)
    return numeric_factor.solve(rhs, inplace=inplace)



#cython: language_level=3str

import warnings
from enum import Enum

import numpy as np
import scipy as sp

from .csc cimport _CSCMatrix
from .csc import _CSCMatrix

cdef extern from "spral.h":
    struct spral_scaling_auction_options:
        int array_base
        int max_iterations
        int max_unchanged[3]
        float min_proportion[3]
        float eps_initial
        char unused[80]

    struct spral_scaling_auction_inform:
        int flag
        int stat
        int matched
        int iterations
        int unmatchable
        char unused[80]

    struct spral_scaling_equilib_options:
        int array_base
        int max_iterations
        float tol
        char unused[80]

    struct spral_scaling_equilib_inform:
        int flag
        int stat
        int iterations
        char unused[80]

    struct spral_scaling_hungarian_options:
        int array_base
        bint scale_if_singular
        char unused[80]

    struct spral_scaling_hungarian_inform:
        int flag
        int stat
        int matched
        char unused[80]

    void spral_scaling_auction_default_options(spral_scaling_auction_options *options)
    void spral_scaling_equilib_default_options(spral_scaling_equilib_options *options)
    void spral_scaling_hungarian_default_options(spral_scaling_hungarian_options *options)

    void spral_scaling_auction_sym(int n, const int *ptr, const int *row,
                                   const double *val, double *scaling, int *match,
                                   const spral_scaling_auction_options *options,
                                   spral_scaling_auction_inform *inform)

    void spral_scaling_equilib_sym(int n, const int *ptr, const int *row,
                                   const double *val, double *scaling,
                                   const spral_scaling_equilib_options *options,
                                   spral_scaling_equilib_inform *inform)

    void spral_scaling_hungarian_sym(int n, const int *ptr, const int *row,
                                     const double *val, double *scaling, int *match,
                                     const spral_scaling_hungarian_options *options,
                                     spral_scaling_hungarian_inform *inform)

    void spral_scaling_auction_unsym(int m, int n, const int *ptr,
                                     const int *row, const double *val, double *rscaling, double *cscaling,
                                     int *match, const spral_scaling_auction_options *options,
                                     spral_scaling_auction_inform *inform)

    void spral_scaling_equilib_unsym(int m, int n, const int *ptr,
                                     const int *row, const double *val, double *rscaling, double *cscaling,
                                     const spral_scaling_equilib_options *options,
                                     spral_scaling_equilib_inform *inform)

    void spral_scaling_hungarian_unsym(int m, int n, const int *ptr,
                                       const int *row, const double *val, double *rscaling, double *cscaling,
                                       int *match, const spral_scaling_hungarian_options *options,
                                       spral_scaling_hungarian_inform *inform)


class ExitCode(Enum):
    Success = 0
    AllocationError = -1
    RankDeficientError = -2
    RankDeficientWarning = 1

    def error(self):
        return self.value < 0

    def warning(self):
        return self.value > 0

    def success(self):
        return self.value == 0

    def __str__(self):
        return "{0} (exit code: {1})".format(code_messages[self], self.value)


code_messages = {
    ExitCode.Success: "Success",
    ExitCode.AllocationError: "Allocation error",
    ExitCode.RankDeficientError: "Rank deficient matrix error",
    ExitCode.RankDeficientWarning: "Rank deficient matrix warning"
}


class ScalingError(Exception):
    def __init__(self, error_code):
        self.error_code = error_code


cdef class AuctionOptions:
    cdef spral_scaling_auction_options options

    def __cinit__(self, **kwargs):
        spral_scaling_auction_default_options(&self.options)

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    property array_base:
        def __get__(self): return self.options.array_base
        def __set__(self, value): self.options.array_base = value

    property max_iterations:
        def __get__(self): return self.options.max_iterations
        def __set__(self, value): self.options.max_iterations = value

    property max_unchanged:
        def __get__(self): return self.options.max_unchanged
        def __set__(self, value): self.options.max_unchanged = value

    property min_proportion:
        def __get__(self): return self.options.min_proportion
        def __set__(self, value): self.options.min_proportion = value


cdef class HungarianOptions:
    cdef spral_scaling_hungarian_options options

    def __cinit__(self, **kwargs):
        spral_scaling_hungarian_default_options(&self.options)

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    property array_base:
        def __get__(self): return self.options.array_base
        def __set__(self, value): self.options.array_base = value

    property scale_if_singular:
        def __get__(self): return self.options.scale_if_singular
        def __set__(self, value): self.options.scale_if_singular = value


cdef class EquilibOptions:
    cdef spral_scaling_equilib_options options

    def __cinit__(self, **kwargs):
        spral_scaling_equilib_default_options(&self.options)

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    property array_base:
        def __get__(self): return self.options.array_base
        def __set__(self, value): self.options.array_base = value

    property max_iterations:
        def __get__(self): return self.options.max_iterations
        def __set__(self, value): self.options.max_iterations = value

    property tol:
        def __get__(self): return self.options.tol
        def __set__(self, value): self.options.tol = value


def raise_error_from_flag(flag):
    exit_code = ExitCode(flag)
    if exit_code.error():
        raise ScalingError(exit_code)
    elif exit_code.warning():
        warnings.warn(code_messages[exit_code], RuntimeWarning)


class SymmetricScalingResult:
    def __init__(self, scaling, matching=None, **props):
        self.scaling = scaling
        self.matching = matching
        self.props = props

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return self.props[name]


class UnsymmetricScalingResult:
    def __init__(self, row_scaling, col_scaling, matching=None, **props):
        self.row_scaling = row_scaling
        self.col_scaling = col_scaling
        self.matching = matching
        self.props = props

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return self.props[name]


cdef _auction_props(spral_scaling_auction_inform* inform):
    return {
        "matched": inform.matched,
        "iterations": inform.iterations,
        "unmatchable": inform.unmatchable
    }


cdef _scaling_auction_sym(A: sp.sparse.spmatrix,
                          AuctionOptions options):
    cdef int m, n
    cdef spral_scaling_auction_inform inform
    cdef _CSCMatrix csc_mat
    cdef double[::1] scaling
    cdef int[::1] matching

    csc_mat = _CSCMatrix(A, symmetric=True)

    (m, n) = csc_mat.shape
    assert m == n, "Matrix must be square"

    options.array_base = 1

    scaling = np.zeros(n, dtype=np.float64)
    matching = np.zeros(n, dtype=np.intc)

    spral_scaling_auction_sym(n,
                              &csc_mat.cols[0],
                              &csc_mat.rows[0],
                              &csc_mat.data[0],
                              &scaling[0],
                              &matching[0],
                              &options.options,
                              &inform)

    raise_error_from_flag(inform.flag)

    return SymmetricScalingResult(np.asarray(scaling),
                                  np.asarray(matching),
                                  **_auction_props(&inform))


cdef _scaling_auction_unsym(A: sp.sparse.spmatrix,
                             AuctionOptions options):
    cdef int m, n
    cdef spral_scaling_auction_inform inform
    cdef _CSCMatrix csc_mat
    cdef double[::1] row_scaling
    cdef double[::1] col_scaling
    cdef int[::1] matching

    csc_mat = _CSCMatrix(A, symmetric=False)

    (m, n) = csc_mat.shape
    options.array_base = 1

    row_scaling = np.zeros(m, dtype=np.float64)
    col_scaling = np.zeros(n, dtype=np.float64)
    matching = np.zeros(n, dtype=np.intc)

    spral_scaling_auction_unsym(m,
                                n,
                                &csc_mat.cols[0],
                                &csc_mat.rows[0],
                                &csc_mat.data[0],
                                &row_scaling[0],
                                &col_scaling[0],
                                &matching[0],
                                &options.options,
                                &inform)

    raise_error_from_flag(inform.flag)

    return UnsymmetricScalingResult(np.asarray(row_scaling),
                                    np.asarray(col_scaling),
                                    np.asarray(matching),
                                    **_auction_props(&inform))


def scaling_auction(A: sp.sparse.spmatrix,
                    symmetric=False,
                    **options):
    cdef AuctionOptions opts = AuctionOptions(**options)

    assert sp.sparse.issparse(A)

    if symmetric:
        return _scaling_auction_sym(A, opts)
    else:
        return _scaling_auction_unsym(A, opts)


cdef _scaling_hungarian_sym(A: sp.sparse.spmatrix,
                            HungarianOptions options):
    cdef int m, n
    cdef spral_scaling_hungarian_inform inform
    cdef _CSCMatrix csc_mat
    cdef double[::1] scaling
    cdef int[::1] matching

    csc_mat = _CSCMatrix(A, symmetric=True)

    (m, n) = csc_mat.shape
    assert m == n, "Matrix must be square"

    options.array_base = 1

    scaling = np.zeros(n, dtype=np.float64)
    matching = np.zeros(n, dtype=np.intc)

    spral_scaling_hungarian_sym(n,
                                &csc_mat.cols[0],
                                &csc_mat.rows[0],
                                &csc_mat.data[0],
                                &scaling[0],
                                &matching[0],
                                &options.options,
                                &inform)

    raise_error_from_flag(inform.flag)

    return SymmetricScalingResult(np.asarray(scaling),
                                  np.asarray(matching),
                                  matched=inform.matched)


cdef _scaling_hungarian_unsym(A: sp.sparse.spmatrix,
                              HungarianOptions options):
    cdef int m, n
    cdef spral_scaling_hungarian_inform inform
    cdef _CSCMatrix csc_mat
    cdef double[::1] row_scaling
    cdef double[::1] col_scaling
    cdef int[::1] matching

    csc_mat = _CSCMatrix(A, symmetric=False)

    (m, n) = csc_mat.shape
    options.array_base = 1

    row_scaling = np.zeros(m, dtype=np.float64)
    col_scaling = np.zeros(n, dtype=np.float64)
    matching = np.zeros(n, dtype=np.intc)

    spral_scaling_hungarian_unsym(m,
                                  n,
                                  &csc_mat.cols[0],
                                  &csc_mat.rows[0],
                                  &csc_mat.data[0],
                                  &row_scaling[0],
                                  &col_scaling[0],
                                  &matching[0],
                                  &options.options,
                                  &inform)

    raise_error_from_flag(inform.flag)

    return UnsymmetricScalingResult(np.asarray(row_scaling),
                                    np.asarray(col_scaling),
                                    np.asarray(matching),
                                    matched=inform.matched)


def scaling_hungarian(A: sp.sparse.spmatrix,
                      symmetric=False,
                      **options):
    cdef HungarianOptions opts = HungarianOptions(**options)

    assert sp.sparse.issparse(A)

    if symmetric:
        return _scaling_hungarian_sym(A, opts)
    else:
        return _scaling_hungarian_unsym(A, opts)


cdef _scaling_equilib_sym(A: sp.sparse.spmatrix,
                          EquilibOptions options):
    cdef int m, n
    cdef spral_scaling_equilib_inform inform
    cdef _CSCMatrix csc_mat
    cdef double[::1] scaling

    csc_mat = _CSCMatrix(A, symmetric=True)

    (m, n) = csc_mat.shape
    assert m == n, "Matrix must be square"

    options.array_base = 1

    scaling = np.zeros(n, dtype=np.float64)

    spral_scaling_equilib_sym(n,
                              &csc_mat.cols[0],
                              &csc_mat.rows[0],
                              &csc_mat.data[0],
                              &scaling[0],
                              &options.options,
                              &inform)

    raise_error_from_flag(inform.flag)

    return SymmetricScalingResult(np.asarray(scaling),
                                  iterations=inform.iterations)


cdef _scaling_equilib_unsym(A: sp.sparse.spmatrix,
                            EquilibOptions options):
    cdef int m, n
    cdef spral_scaling_equilib_inform inform
    cdef _CSCMatrix csc_mat
    cdef double[::1] row_scaling
    cdef double[::1] col_scaling
    cdef int[::1] matching

    csc_mat = _CSCMatrix(A, symmetric=False)

    (m, n) = csc_mat.shape
    options.array_base = 1

    row_scaling = np.zeros(m, dtype=np.float64)
    col_scaling = np.zeros(n, dtype=np.float64)

    spral_scaling_equilib_unsym(m,
                                n,
                                &csc_mat.cols[0],
                                &csc_mat.rows[0],
                                &csc_mat.data[0],
                                &row_scaling[0],
                                &col_scaling[0],
                                &options.options,
                                &inform)

    raise_error_from_flag(inform.flag)

    return UnsymmetricScalingResult(np.asarray(row_scaling),
                                    np.asarray(col_scaling),
                                    iterations=inform.iterations)


def scaling_equilib(A: sp.sparse.spmatrix,
                    symmetric=False,
                    **options):
    cdef EquilibOptions opts = EquilibOptions(**options)

    assert sp.sparse.issparse(A)

    if symmetric:
        return _scaling_equilib_sym(A, opts)
    else:
        return _scaling_equilib_unsym(A, opts)


class ScalingMethod(Enum):
    Auction = 0
    Hungarian = 1
    Equilib = 2

    def __str__(self):
        return self.name


def scaling(A: sp.sparse.spmatrix,
            method: ScalingMethod,
            symmetric=False,
            **options):
    if method == ScalingMethod.Auction:
        return scaling_auction(A, symmetric, **options)
    elif method == ScalingMethod.Hungarian:
        return scaling_hungarian(A, symmetric, **options)
    elif method == ScalingMethod.Equilib:
        return scaling_equilib(A, symmetric, **options)
    else:
        raise ValueError("Unknown scaling method")



from importlib.resources import files
import tempfile
import os

import numpy as np
import scipy as sp
import pytest

from pyspral.rb import peek, read, write, Matrix, MatrixType, LowerUpperValues, ReadValues
import tests.resources


@pytest.fixture()
def mat():
    _mat = np.array([[ 1.0, 0.0, -1.0 ],
                     [ 0.0, 2.0,  0.0 ],
                     [ 2.0, 0.0,  2.0 ],
                     [ 5.0, 3.0, -2.0 ],
                     [ 0.0, 0.0,  6.0 ]])

    return sp.sparse.csc_matrix(np.copy(_mat))


@pytest.fixture
def rb_file_name():
    return files(tests.resources).joinpath("matrix.rb")


def test_peek(rb_file_name):
    result = peek(rb_file_name)

    assert result.shape == (5, 5)

    assert result.matrix_type == MatrixType.RealSymmIndef


def test_read(rb_file_name):
    mat = read(rb_file_name, lwr_upr_full=LowerUpperValues.Lower)

    assert mat.shape == (5, 5)

    mat_val = mat.tocsc()
    assert mat_val.shape == (5, 5)

    matcoo = mat_val.tocoo()

    assert (matcoo.row >= matcoo.col).all()


def check_symm(mat):
    mat_val = mat.tocsc()
    matcoo = mat_val.tocoo()

    row = matcoo.row
    col = matcoo.col
    data = matcoo.data

    entries = dict()

    for (i, j, d) in zip(row, col, data):
        entries[(i, j)] = d

    for (i, j, d) in zip(row, col, data):
        if i == j:
            continue

        assert (j, i) in entries

        assert entries[(i, j)] == entries[(j, i)]


def test_read_full(rb_file_name):
    mat = read(rb_file_name, lwr_upr_full=LowerUpperValues.Full)

    assert mat.shape == (5, 5)

    check_symm(mat)


def test_read_random(rb_file_name):
    mat = read(rb_file_name,
               random_state=42,
               lwr_upr_full=LowerUpperValues.Full,
               values=ReadValues.RandSymm)

    next_mat = read(rb_file_name,
                    random_state=42,
                    lwr_upr_full=LowerUpperValues.Full,
                    values=ReadValues.RandSymm)

    assert (mat.data == next_mat.data).all()


def test_write_mat(mat):
    title = "Test Matrix"
    identifier = "TestId"

    matrix = Matrix.fromcsc(mat,
                            title=title,
                            identifier=identifier,
                            matrix_type=MatrixType.RealRect)

    tmp_dir = None

    try:
        tmp_dir = tempfile.TemporaryDirectory()

        filename = os.path.join(tmp_dir.name, "matrix.rb")

        write(filename, matrix)

        other_matrix = read(filename)

        assert (other_matrix.tocsc() - mat).nnz == 0

        assert other_matrix.title == title
        assert other_matrix.identifier == identifier
    finally:
        if tmp_dir is not None:
            tmp_dir.cleanup()



#cython: language_level=3str

from enum import Enum

cdef extern from "spral.h":
    enum spral_matrix_type:
        SPRAL_MATRIX_UNSPECIFIED = 0
        SPRAL_MATRIX_REAL_RECT = 1
        SPRAL_MATRIX_CPLX_RECT = -1
        SPRAL_MATRIX_REAL_UNSYM = 2
        SPRAL_MATRIX_CPLX_UNSYM = -2
        SPRAL_MATRIX_REAL_SYM_PSDEF = 3
        SPRAL_MATRIX_CPLX_HERM_PSDEF = -3
        SPRAL_MATRIX_REAL_SYM_INDEF = 4
        SPRAL_MATRIX_CPLX_HERM_INDEF = -4
        SPRAL_MATRIX_CPLX_SYM = -5
        SPRAL_MATRIX_REAL_SKEW = 6
        SPRAL_MATRIX_CPLX_SKEW = -6



import scipy as sp
import numpy as np

import pytest

from pyspral.lsmr import solve


@pytest.fixture()
def mat():
    _mat = np.array([[ 1.0, 0.0, -1.0 ],
                     [ 0.0, 2.0,  0.0 ],
                     [ 2.0, 0.0,  2.0 ],
                     [ 5.0, 3.0, -2.0 ],
                     [ 0.0, 0.0,  6.0 ]])

    return sp.sparse.linalg.aslinearoperator(np.copy(_mat))


@pytest.fixture()
def rhs():
    return np.array([1.0, 1.0, 1.0, 1.0, 1.0])


def test_solve(mat, rhs):
    result = solve(mat, rhs)

    assert result.success


def test_solve_damp(mat, rhs):
    result = solve(mat, rhs, lamb=1.)

    assert result.success


def test_solve_options(mat, rhs):
    result = solve(mat, rhs, itnlim=1)

    assert not result.success
    assert result.status == "iter_limit"


def test_solve_precond(mat, rhs):
    (_, n) = mat.shape
    precond = sp.sparse.linalg.aslinearoperator(np.eye(n))

    result = solve(mat, rhs, P=precond)

    assert result.success


def test_simple_callback(mat, rhs):

    def callback(x, iteration):
        return False

    result = solve(mat,
                   rhs,
                   callback=callback,
                   ctest="manual_simple")

    assert result.status == "manual_abort"


def test_callback(mat, rhs):

    all_norms = []

    def callback(x, iteration, norms):
        all_norms.append(norms["normAPr"])
        return True

    result = solve(mat,
                   rhs,
                   callback=callback,
                   ctest="manual_norm")

    assert len(all_norms) > 0
    assert np.isclose(all_norms[-1], 0.)



#cython: language_level=3

from setuptools import Extension, setup
from Cython.Build import cythonize

import numpy as np


modules = ["csc",
           "lsmr",
           "random_matrix",
           "rb",
           "scaling",
           "ssids",
           "ssmfe",
           "types"]


extensions = [
    Extension(
        "pyspral.{0}".format(module),
        sources=["src/pyspral/{0}.pyx".format(module)],
        libraries=["spral"],
        include_dirs=[np.get_include()]
    )
    for module in modules
]

setup(ext_modules=cythonize(extensions, gdb_debug=True))



#cython: language_level=3str

from enum import Enum
import warnings
from typing import Optional

cimport numpy as np

import numpy as np
import scipy as sp


cdef extern from "spral.h":

    struct spral_ssmfe_rcid:
        int job
        int nx
        int jx
        int kx
        int ny
        int jy
        int ky
        int i
        int j
        int k
        double alpha
        double beta
        double *x
        double *y
        char unused[80]

    struct spral_ssmfe_rciz:
        int job
        int nx
        int jx
        int kx
        int ny
        int jy
        int ky
        int i
        int j
        int k
        np.complex128_t alpha
        np.complex128_t beta
        np.complex128_t *x
        np.complex128_t *y
        char unused[80]

    struct spral_ssmfe_core_options:
        int array_base
        double cf_max
        int err_est
        int extra_left
        int extra_right
        double min_gap
        bint minAprod
        bint minBprod
        char unused[80]

    struct spral_ssmfe_options:
        int array_base
        int print_level
        int unit_error
        int unit_warning
        int unit_diagnostic
        int max_iterations
        int user_x
        int err_est
        double abs_tol_lambda
        double rel_tol_lambda
        double abs_tol_residual
        double rel_tol_residual
        double tol_x
        double left_gap
        double right_gap
        int extra_left
        int extra_right
        int max_left
        int max_right
        bint minAprod
        bint minBprod

    struct spral_ssmfe_inform:
        int flag
        int stat
        int non_converged
        int iteration
        int left
        int right
        int *converged
        double next_left
        double next_right
        double *residual_norms
        double *err_lambda
        double *err_X
        char unused[80]

    void spral_ssmfe_default_options(spral_ssmfe_options *options);

    void spral_ssmfe_standard_double(spral_ssmfe_rcid *rci, int left,
          int mep, double *lamb, int n, double *x, int ldx,
          void **keep, const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_standard_double_complex(spral_ssmfe_rciz *rci,
          int left, int mep, double *lamb, int n, np.complex128_t *x,
          int ldx, void **keep, const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_standard_shift_double(spral_ssmfe_rcid *rci,
          double sigma, int left, int right, int mep, double *lamb, int n,
          double *x, int ldx, void **keep,
          const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_standard_shift_double_complex(
          spral_ssmfe_rciz *rci, double sigma, int left, int right, int mep,
          double *lamb, int n, np.complex128_t *x, int ldx, void **keep,
          const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_generalized_double(spral_ssmfe_rcid *rci,
          int left, int mep, double *lamb, int n, double *x, int ldx,
          void **keep, const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_generalized_double_complex(spral_ssmfe_rciz *rci,
          int left, int mep, double *lamb, int n, np.complex128_t *x,
          int ldx, void **keep, const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_generalized_shift_double(spral_ssmfe_rcid *rci,
          double sigma, int left, int right, int mep, double *lamb, int n,
          double *x, int ldx, void **keep,
          const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_generalized_shift_double_complex(
          spral_ssmfe_rciz *rci, double sigma, int left, int right, int mep,
          double *lamb, int n, np.complex128_t *x, int ldx, void **keep,
          const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_buckling_double(spral_ssmfe_rcid *rci,
          double sigma, int left, int right, int mep, double *lamb, int n,
          double *x, int ldx, void **keep,
          const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_buckling_double_complex(spral_ssmfe_rciz *rci,
          double sigma, int left, int right, int mep, double *lamb, int n,
          np.complex128_t *x, int ldx, void **keep,
          const spral_ssmfe_options *options,
          spral_ssmfe_inform *inform)

    void spral_ssmfe_free_double(void **keep, spral_ssmfe_inform *inform)

    void spral_ssmfe_free_double_complex(void **keep, spral_ssmfe_inform *inform)


class Flag(Enum):
    JOB_OUT_OF_RANGE = -1
    N_OUT_OF_RANGE = -9
    LDX_OUT_OF_RANGE = -10
    LEFT_OUT_OF_RANGE = -11
    RIGHT_OUT_OF_RANGE = -12
    MEP_LT_N_EIGENPAIRS = -13
    NOT_ENOUGH_MEM = -100
    MAT_NOT_POSDEF = -200
    COMPUTATION_FINISHED = 1
    EXCEEDED_MAX_NUM_ITS = 2
    OUT_OF_STORAGE_FOR_EIGENPAIRS = 3


class Result:
    def __init__(self, lamb, x, **props):
        self.lamb = lamb
        self.x = x
        self.props = props

    def __getattr__(self, name):
        return self.props[name]


class PrintLevel(Enum):
    NoPrinting = -1
    ErrorWarning = 0
    BasicDiagnostic = 1
    DetailedDiagnostic = 2


cdef class Options:
    cdef spral_ssmfe_options options

    def __cinit__(self, **options):
        spral_ssmfe_default_options(&self.options)

    def __init__(self, **options):
        for key, value in options.items():
            setattr(self, key, value)

    property array_base:
        def __set__(self, value): self.options.array_base = value

    property print_level:
        def __set__(self, value): self.options.print_level = value.value

    property unit_error:
        def __set__(self, value): self.options.unit_error = value

    property unit_warning:
        def __set__(self, value): self.options.unit_warning = value

    property unit_diagnostic:
        def __set__(self, value): self.options.unit_diagnostic = value

    property max_iterations:
        def __set__(self, value): self.options.max_iterations = value

    property user_x:
        def __set__(self, value): self.options.user_x = value

    property err_est:
        def __set__(self, value): self.options.err_est = value

    property abs_tol_lambda:
        def __set__(self, value): self.options.abs_tol_lambda = value

    property rel_tol_lambda:
        def __set__(self, value): self.options.rel_tol_lambda = value

    property abs_tol_residual:
        def __set__(self, value): self.options.abs_tol_residual = value

    property rel_tol_residual:
        def __set__(self, value): self.options.rel_tol_residual = value

    property tol_x:
        def __set__(self, value): self.options.tol_x = value

    property left_gap:
        def __set__(self, value): self.options.left_gap = value

    property right_gap:
        def __set__(self, value): self.options.right_gap = value

    property extra_left:
        def __set__(self, value): self.options.extra_left = value

    property extra_right:
        def __set__(self, value): self.options.extra_right = value

    property max_left:
        def __set__(self, value): self.options.max_left = value

    property max_right:
        def __set__(self, value): self.options.max_right = value

    property minAprod:
        def __set__(self, value): self.options.minAprod = value

    property minBprod:
        def __set__(self, value): self.options.minBprod = value


cdef _solve_standard(A: sp.sparse.linalg.LinearOperator,
                     left: int,
                     mep: int,
                     Options options,
                     P: Optional[sp.sparse.linalg.LinearOperator] = None,
                     x0: double[:,::1]=None):
    cdef spral_ssmfe_inform inform
    cdef void* keep = NULL

    cdef int _left = left
    cdef int _mep = mep
    cdef int n
    cdef int ldx
    cdef double[:] lamb
    cdef double[:, ::1] x
    cdef spral_ssmfe_rcid rci

    assert mep >= left

    rci.job = 0

    (m, n) = A.shape

    if m != n:
        raise ValueError("Matrix A is not square")

    lamb = np.zeros(left, dtype=np.float64)

    if x0 is None:
        x = np.zeros((mep, n), dtype=np.float64)
    else:
        assert x0.shape == (mep, n)
        x = x0

    ldx = n

    options.left_gap = -0.1

    try:
        while True:
            spral_ssmfe_standard_double(&rci,
                                        _left,
                                        _mep,
                                        &lamb[0],
                                        n,
                                        &x[0, 0],
                                        ldx,
                                        &keep,
                                        &options.options,
                                        &inform)

            if rci.job == -3:
                # Fatal error
                raise ValueError("Fatal error")
            elif rci.job == -2:
                # Error, restart?
                raise ValueError("Error occurred during computation")
            elif rci.job == -1:
                # Computation complete
                break
            elif rci.job == 1:
                # Compute Y = AX
                xv = np.asarray(<double[:rci.nx, :n]> rci.x)
                yv = np.asarray(<double[:rci.nx, :n]> rci.y)
                yv[:] = (A @ xv.T).T
            elif rci.job == 2:
                # Apply precond Y = TX
                xv = np.asarray(<double[:rci.nx, :n]> rci.x)
                yv = np.asarray(<double[:rci.nx, :n]> rci.y)

                if P is None:
                    yv[:] = xv
                else:
                    yv[:] = (P @ xv.T).T
                continue
            else:
                raise ValueError("Invalid job")
    finally:
        spral_ssmfe_free_double(&keep, &inform)

    num_eigenpairs = inform.left

    props = {
        'iteration': inform.iteration,
        'next_left': inform.next_left,
    }

    if num_eigenpairs > 0:
        lambv = np.asarray(lamb[:num_eigenpairs])
        xv = np.asarray(x[:num_eigenpairs, :]).T
    else:
        lambv = np.empty(shape=(0,), dtype=np.float64)
        xv = np.empty(shape=(n, 0), dtype=np.float64)

    return Result(lambv, xv, **props)


def solve_standard(A: sp.sparse.linalg.LinearOperator,
                   left: int,
                   mep: int,
                   P: Optional[sp.sparse.linalg.LinearOperator] = None,
                   x0: double[:,::1]=None,
                   **options):
    cdef Options opts = Options(**options)
    return _solve_standard(A, left, mep, opts, P, x0)


cdef _solve_generalized(A: sp.sparse.linalg.LinearOperator,
                        B: sp.sparse.linalg.LinearOperator,
                        left: int,
                        mep: int,
                        Options options,
                        P: Optional[sp.sparse.linalg.LinearOperator] = None,
                        x0: double[:,::1]=None):
    cdef spral_ssmfe_inform inform
    cdef void* keep = NULL

    cdef int _left = left
    cdef int _mep = mep
    cdef int n
    cdef int ldx
    cdef double[:] lamb
    cdef double[:, ::1] x
    cdef spral_ssmfe_rcid rci

    assert mep >= left

    rci.job = 0

    (m, n) = A.shape

    if m != n:
        raise ValueError("Matrix A is not square")

    lamb = np.zeros(left, dtype=np.float64)

    if x0 is None:
        x = np.zeros((mep, n), dtype=np.float64)
    else:
        assert x0.shape == (mep, n)
        x = x0

    ldx = n

    try:
        while True:
            spral_ssmfe_generalized_double(&rci,
                                           _left,
                                           _mep,
                                           &lamb[0],
                                           n,
                                           &x[0, 0],
                                           ldx,
                                           &keep,
                                           &options.options,
                                           &inform)

            if rci.job == -3:
                # Fatal error
                raise ValueError("Fatal error")
            elif rci.job == -2:
                # Error, restart?
                raise ValueError("Error occurred during computation")
            elif rci.job == -1:
                # Computation complete
                break
            elif rci.job == 1:
                # Compute Y = AX
                xv = np.asarray(<double[:rci.nx, :n]> rci.x)
                yv = np.asarray(<double[:rci.nx, :n]> rci.y)
                yv[:] = (A @ xv.T).T
            elif rci.job == 2:
                # Apply precond Y = TX
                xv = np.asarray(<double[:rci.nx, :n]> rci.x)
                yv = np.asarray(<double[:rci.nx, :n]> rci.y)

                if P is None:
                    yv[:] = xv
                else:
                    yv[:] = (P @ xv.T).T
                continue

            elif rci.job == 3:
                # Compute Y = BX
                xv = np.asarray(<double[:rci.nx, :n]> rci.x)
                yv = np.asarray(<double[:rci.nx, :n]> rci.y)

                yv[:] = (B @ xv.T).T
            else:
                raise ValueError("Invalid job")
    finally:
        spral_ssmfe_free_double(&keep, &inform)

    num_eigenpairs = inform.left

    props = {
        'iteration': inform.iteration,
        'next_left': inform.next_left,
    }

    if num_eigenpairs > 0:
        lambv = np.asarray(lamb[:num_eigenpairs])
        xv = np.asarray(x[:num_eigenpairs, :]).T
    else:
        lambv = np.empty(shape=(0,), dtype=np.float64)
        xv = np.empty(shape=(n, 0), dtype=np.float64)

    return Result(lambv, xv, **props)


def solve_generalized(A: sp.sparse.linalg.LinearOperator,
                      B: sp.sparse.linalg.LinearOperator,
                      left: int,
                      mep: int,
                      P: Optional[sp.sparse.linalg.LinearOperator] = None,
                      x0: double[:,::1]=None,
                      **options):
    cdef Options opts = Options(**options)
    return _solve_generalized(A, B, left, mep, opts, P, x0)


cpdef solve_buckling():
    raise NotImplementedError()



#cython: language_level=3str

from enum import Flag, Enum
from libc.stdint cimport int64_t

cimport numpy as np
import numpy as np
import scipy as sp

from .types cimport spral_matrix_type
from .types import MatrixType


cdef extern from "spral.h":
    int spral_random_matrix_generate(int *state,
                                     spral_matrix_type matrix_type,
                                     int m,
                                     int n,
                                     int nnz,
                                     int *ptr,
                                     int *row,
                                     double *val,
                                     int flags)

    int spral_random_matrix_generate_long(int *state,
                                          spral_matrix_type matrix_type,
                                          int m,
                                          int n,
                                          int64_t nnz,
                                          int64_t *ptr,
                                          int *row,
                                          double *val,
                                          int flags)



class MatrixFlags(Flag):
    ZERO = 0
    FINDEX = 1
    NONSINGULAR = 2
    SORTED = 4


class StatusCode(Enum):
    Success = 0
    AllocationError = -1
    InvalidMatrixType = -2
    NonPositiveValue = -3
    InconsistentDimensions = -4
    InsuffiicentNNZ = -5

    def success(self):
        return self == StatusCode.Success


class Error(Exception):
    def __init__(self, code):
        self.code = code

    def __str__(self):
        return "Failed to generate random matrix: {0}".format(self.code)


def random_matrix_generate(m,
                           n,
                           nnz,
                           matrix_type,
                           bint nonsingular=False,
                           int random_state=0):

    if nnz <= 0:
        raise ValueError("Invalid number of nonzero values provided")

    if n <= 0 or m <= 0:
        raise ValueError("Invalid dimensions provided")

    cdef int ptr_size = n + 1
    cdef int[:] ptr = np.empty((ptr_size,), dtype=np.intc)
    cdef int[:] row = np.empty((nnz,), dtype=np.intc)
    cdef double[:] val = np.empty((nnz,), dtype=float)
    cdef int result
    cdef int flag
    cdef spral_matrix_type mat_type = matrix_type.value

    flags = MatrixFlags.SORTED

    if nonsingular:
        flags = flags | MatrixFlags.NONSINGULAR

    flag = flags.value

    result = spral_random_matrix_generate(&random_state,
                                          mat_type,
                                          m,
                                          n,
                                          nnz,
                                          &ptr[0],
                                          &row[0],
                                          &val[0],
                                          flag)

    code = StatusCode(result)

    if not code.success():
        raise Error(code)

    return sp.sparse.csc_matrix((np.asarray(val),
                                 np.asarray(row),
                                 np.asarray(ptr)),
                                shape=(m, n))
